#include <predef.h>
#include <constants.h>
#include <ctype.h>
#include <ucos.h>
#include <system.h>
#include <init.h>
#include <smarttrap.h>
#include <utils.h>
#include <http.h>
#include <iosys.h>
#include "jpeg-compressor/jpge.h"

#define __PACKED __attribute__ ((__packed__))

const char AppName[] = "bmp2jpeg";

extern "C" {
    void UserMain( void * pd );
}


uint8_t * lena_dat;

namespace tiff {

enum ifd_type {
    BYTE    = 1,
    ASCII   = 2,
    SHORT   = 3,
    LONG    = 4,
    RATIONAL= 5,
    SBYTE   = 6,
    UNDEFINED=7,
    SSHORT  = 8,
    SLONG   = 9,
    SRATIONAL=10,
    FLOAT   = 11,
    DOUBLE  = 12,
    TYPE_16BIT = 0xFFFF
};

enum ifd_tag {
    NEW_SUB_FILETYPE = 254,
    SUB_FILETYPE = 255,
    IMAGE_WIDTH = 256,
    IMAGE_LENGTH = 257,
    BITS_PER_SAMPLE = 258,
    COMPRESSION = 259,
    PHOTOMETRIC_INTERPRETATION = 262,
    THRESHHOLDING = 263,
    CELL_WIDTH = 264,
    CELL_LENGTH = 265,
    FILL_ORDER = 266,
    DOCUMENT_NAME = 269,
    IMAGE_DESCRIPTION = 270,
    MAKE = 271,
    MODEL = 272,
    STRIP_OFFSETS = 273,
    ORIENTATION = 274,
    SAMPLES_PER_PIXEL = 277,
    ROWS_PER_STRIP = 278,
    STRIP_BYTE_COUNTS = 279,
    MIN_SAMPLE_VALUE = 280,
    MAX_SAMPLE_VALUE = 281,
    X_RESOLUTION = 282,
    Y_RESOLUTION = 283,
    PLANAR_CONFIGURATION = 284,
    PAGE_NAME = 285,
    X_POSITION = 286,
    Y_POSITION = 287,
    FREE_OFFSETS = 288,
    FREE_BYTE_COUNTS = 289,
    GRAY_RESPONSE_UNIT = 290,
    GRAY_RESPONSE_CURVE = 291,
    T4_OPTIONS = 292,
    T6_OPTIONS = 293,
    RESOLUTION_UNIT = 296,
    PAGE_NUMBER = 297,
    TRANSFER_FUNCTION = 301,
    SOFTWARE = 305,
    DATETIME = 306,
    ARTIST = 315,
    HOST_COMPUTER = 316,
    PREDICTOR = 317,
    WHITE_POINT = 318,
    PRIMARY_CHROMATICITIES = 319,
    COLORMAP = 320,
    HALF_TONE_HINTS = 321,
    TILE_WIDTH = 322,
    TILE_LENGTH = 323,
    TILE_OFFSETS = 324,
    TILE_BYTE_COUNTS = 325,
    INK_SET = 332,
    INK_NAMES = 333,
    NUMBER_OF_INKS = 334,
    DOT_RANGE = 336,
    TARGET_PRINTER = 337,
    EXTRA_SAMPLES = 338,
    SAMPLE_FORMAT = 339,
    S_MIN_SAMPLE_VALUE = 340,
    S_MAX_SAMPLE_VALUE = 341,
    TRANSFER_RANGE = 342,
    JPEG_PROC = 512,
    JPEG_INTERCHANGE_FORMAT = 513,
    JPEG_INTERCHANGE_FORMAT_LNGTH = 514,
    JPEG_RESTART_INTERVAL = 515,
    JPEG_LOSSLESS_PREDICTORS = 517,
    JPEG_POINT_TRANSFORMS = 518,
    J_PEGQ_TABLES = 519,
    J_PEGDC_TABLES = 520,
    J_PEGAC_TABLES = 521,
    YCBCR_COEFFICIENTS = 529,
    YCBCR_SUBSAMPLING = 530,
    YCBCR_POSITIONING = 531,
    REFERENCE_BLACK_WHITE = 532,
    COPYRIGHT = 33432,
    TAG_16BIT = 0xFFFF
};

struct ifd_entry {
    uint16_t tag;
    uint16_t type;
    uint32_t count;
    uint32_t offset;
    ifd_entry * GetNext() { return &(this[1]); }
} __PACKED;

struct ifd_struct {
    uint16_t count;
    ifd_entry entry[];
} __PACKED;

struct header {
    uint16_t order;
    uint16_t fortytwo;
    uint32_t offset;
    uint8_t  data[];

    inline uint8_t *operator[](uint32_t index) { return data-sizeof(header)+index; }
    inline ifd_struct *GetIFD() {
        return (ifd_struct *)((*this)[offset]);
    }
} __PACKED;

struct Tiff {
    header *head;

    ifd_entry * GetEntry(ifd_tag tag) {
        ifd_struct *ifd = head->GetIFD();
        for (uint16_t i = 0; i < ifd->count; i++) {
            if (ifd->entry[i].tag == tag) { return ifd->entry + i; }
        }
        return NULL;
    }
};


char * GetType(ifd_type type)
{
    switch (type) {
    case BYTE:      { return "BYTE"; }
    case ASCII:     { return "ASCII"; }
    case SHORT:     { return "SHORT"; }
    case LONG:      { return "LONG"; }
    case RATIONAL:  { return "RATIONAL"; }
    case SBYTE:     { return "SBYTE"; }
    case UNDEFINED: { return "UNDEFINED"; }
    case SSHORT:    { return "SSHORT"; }
    case SLONG:     { return "SLONG"; }
    case SRATIONAL: { return "SRATIONAL"; }
    case FLOAT:     { return "FLOAT"; }
    case DOUBLE:    { return "DOUBLE"; }
    case TYPE_16BIT:{ return "TYPE_16BIT"; }
    }
}

char * GetTag(ifd_tag tag)
{
    switch (tag) {
    case NEW_SUB_FILETYPE:              { return "NEW_SUB_FILETYPE"; }
    case SUB_FILETYPE:                  { return "SUB_FILETYPE"; }
    case IMAGE_WIDTH:                   { return "IMAGE_WIDTH"; }
    case IMAGE_LENGTH:                  { return "IMAGE_LENGTH"; }
    case BITS_PER_SAMPLE:               { return "BITS_PER_SAMPLE"; }
    case COMPRESSION:                   { return "COMPRESSION"; }
    case PHOTOMETRIC_INTERPRETATION:    { return "PHOTOMETRIC_INTERPRETATION"; }
    case THRESHHOLDING:                 { return "THRESHHOLDING"; }
    case CELL_WIDTH:                    { return "CELL_WIDTH"; }
    case CELL_LENGTH:                   { return "CELL_LENGTH"; }
    case FILL_ORDER:                    { return "FILL_ORDER"; }
    case DOCUMENT_NAME:                 { return "DOCUMENT_NAME"; }
    case IMAGE_DESCRIPTION:             { return "IMAGE_DESCRIPTION"; }
    case MAKE:                          { return "MAKE"; }
    case MODEL:                         { return "MODEL"; }
    case STRIP_OFFSETS:                 { return "STRIP_OFFSETS"; }
    case ORIENTATION:                   { return "ORIENTATION"; }
    case SAMPLES_PER_PIXEL:             { return "SAMPLES_PER_PIXEL"; }
    case ROWS_PER_STRIP:                { return "ROWS_PER_STRIP"; }
    case STRIP_BYTE_COUNTS:             { return "STRIP_BYTE_COUNTS"; }
    case MIN_SAMPLE_VALUE:              { return "MIN_SAMPLE_VALUE"; }
    case MAX_SAMPLE_VALUE:              { return "MAX_SAMPLE_VALUE"; }
    case X_RESOLUTION:                  { return "X_RESOLUTION"; }
    case Y_RESOLUTION:                  { return "Y_RESOLUTION"; }
    case PLANAR_CONFIGURATION:          { return "PLANAR_CONFIGURATION"; }
    case PAGE_NAME:                     { return "PAGE_NAME"; }
    case X_POSITION:                    { return "X_POSITION"; }
    case Y_POSITION:                    { return "Y_POSITION"; }
    case FREE_OFFSETS:                  { return "FREE_OFFSETS"; }
    case FREE_BYTE_COUNTS:              { return "FREE_BYTE_COUNTS"; }
    case GRAY_RESPONSE_UNIT:            { return "GRAY_RESPONSE_UNIT"; }
    case GRAY_RESPONSE_CURVE:           { return "GRAY_RESPONSE_CURVE"; }
    case T4_OPTIONS:                    { return "T4_OPTIONS"; }
    case T6_OPTIONS:                    { return "T6_OPTIONS"; }
    case RESOLUTION_UNIT:               { return "RESOLUTION_UNIT"; }
    case PAGE_NUMBER:                   { return "PAGE_NUMBER"; }
    case TRANSFER_FUNCTION:             { return "TRANSFER_FUNCTION"; }
    case SOFTWARE:                      { return "SOFTWARE"; }
    case DATETIME:                      { return "DATETIME"; }
    case ARTIST:                        { return "ARTIST"; }
    case HOST_COMPUTER:                 { return "HOST_COMPUTER"; }
    case PREDICTOR:                     { return "PREDICTOR"; }
    case WHITE_POINT:                   { return "WHITE_POINT"; }
    case PRIMARY_CHROMATICITIES:        { return "PRIMARY_CHROMATICITIES"; }
    case COLORMAP:                      { return "COLORMAP"; }
    case HALF_TONE_HINTS:               { return "HALF_TONE_HINTS"; }
    case TILE_WIDTH:                    { return "TILE_WIDTH"; }
    case TILE_LENGTH:                   { return "TILE_LENGTH"; }
    case TILE_OFFSETS:                  { return "TILE_OFFSETS"; }
    case TILE_BYTE_COUNTS:              { return "TILE_BYTE_COUNTS"; }
    case INK_SET:                       { return "INK_SET"; }
    case INK_NAMES:                     { return "INK_NAMES"; }
    case NUMBER_OF_INKS:                { return "NUMBER_OF_INKS"; }
    case DOT_RANGE:                     { return "DOT_RANGE"; }
    case TARGET_PRINTER:                { return "TARGET_PRINTER"; }
    case EXTRA_SAMPLES:                 { return "EXTRA_SAMPLES"; }
    case SAMPLE_FORMAT:                 { return "SAMPLE_FORMAT"; }
    case S_MIN_SAMPLE_VALUE:            { return "S_MIN_SAMPLE_VALUE"; }
    case S_MAX_SAMPLE_VALUE:            { return "S_MAX_SAMPLE_VALUE"; }
    case TRANSFER_RANGE:                { return "TRANSFER_RANGE"; }
    case JPEG_PROC:                     { return "JPEG_PROC"; }
    case JPEG_INTERCHANGE_FORMAT:       { return "JPEG_INTERCHANGE_FORMAT"; }
    case JPEG_INTERCHANGE_FORMAT_LNGTH: { return "JPEG_INTERCHANGE_FORMAT_LNGTH"; }
    case JPEG_RESTART_INTERVAL:         { return "JPEG_RESTART_INTERVAL"; }
    case JPEG_LOSSLESS_PREDICTORS:      { return "JPEG_LOSSLESS_PREDICTORS"; }
    case JPEG_POINT_TRANSFORMS:         { return "JPEG_POINT_TRANSFORMS"; }
    case J_PEGQ_TABLES:                 { return "J_PEGQ_TABLES"; }
    case J_PEGDC_TABLES:                { return "J_PEGDC_TABLES"; }
    case J_PEGAC_TABLES:                { return "J_PEGAC_TABLES"; }
    case YCBCR_COEFFICIENTS:            { return "YCBCR_COEFFICIENTS"; }
    case YCBCR_SUBSAMPLING:             { return "YCBCR_SUBSAMPLING"; }
    case YCBCR_POSITIONING:             { return "YCBCR_POSITIONING"; }
    case REFERENCE_BLACK_WHITE:         { return "REFERENCE_BLACK_WHITE"; }
    case COPYRIGHT:                     { return "COPYRIGHT"; }
    case TAG_16BIT:                     { return "TAG_16BIT"; }
    }
}

void dumpByte(ifd_entry *entry, header &tif)
{
    uint8_t * vals;
    if (entry->count > 4) {
        vals = (uint8_t *) (tif[entry->offset]);
    }
    else {
        vals = (uint8_t*)&entry->offset;
    }
    for (uint16_t i = 0; i < entry->count; i++) {
        iprintf("%02X ", vals[i]);
    }
    putchar('\n');
}

void dumpShort(ifd_entry *entry, header &tif)
{
    uint16_t * vals;
    if (entry->count > 2) {
        vals = (uint16_t *)(tif[entry->offset]);
    }
    else {
        vals = (uint16_t*)&entry->offset;
    }
    for (uint16_t i = 0; i < entry->count; i++) {
        iprintf("%u ", vals[i]);
    }
    putchar('\n');
}

void dumpLong(ifd_entry *entry, header &tif)
{
    uint32_t * vals;
    if (entry->count > 1) {
        vals = (uint32_t *)(tif[entry->offset]);
    }
    else {
        vals = (uint32_t*)&entry->offset;
    }
    for (uint16_t i = 0; i < entry->count; i++) {
        iprintf("%u ", vals[i]);
    }
    putchar('\n');
}

void dumpRational(ifd_entry *entry, header &tif)
{
    uint32_t * vals = (uint32_t *)(tif[entry->offset]);
    for (uint16_t i = 0; i>>1 < entry->count; i+=2) {
        iprintf("%u/%u ", vals[i], vals[i+1]);
    }
    putchar('\n');
}

void dumpSShort(ifd_entry *entry, header &tif)
{
    uint16_t * vals;
    if (entry->count > 2) {
        vals = (uint16_t *)(tif[entry->offset]);
    }
    else {
        vals = (uint16_t*)&entry->offset;
    }
    for (uint16_t i = 0; i < entry->count; i++) {
        iprintf("%d ", vals[i]);
    }
    putchar('\n');
}

void dumpSLong(ifd_entry *entry, header &tif)
{
    uint32_t * vals;
    if (entry->count > 1) {
        vals = (uint32_t *)(tif[entry->offset]);
    }
    else {
        vals = (uint32_t*)&entry->offset;
    }
    for (uint16_t i = 0; i < entry->count; i++) {
        iprintf("%d ", vals[i]);
    }
    putchar('\n');
}

void dumpSRational(ifd_entry *entry, header &tif)
{
    uint32_t * vals = (uint32_t *)(tif[entry->offset]);
    for (uint16_t i = 0; i>>1 < entry->count; i+=2) {
        iprintf("%d/%d ", vals[i], vals[i+1]);
    }
    putchar('\n');
}

void dumpUndef(ifd_entry *entry, header &tif)
{
    iprintf("0x%08X ", entry->offset);
    ShowData(tif[entry->offset], entry->count);
    putchar('\n');
}

void dumpFloat(ifd_entry *entry, header &tif)
{
    float * vals;
    if (entry->count > 1) {
        vals = (float *)(tif[entry->offset]);
    }
    else {
        vals = (float*)&entry->offset;
    }
    for (uint16_t i = 0; i < entry->count; i++) {
        printf("%f ", vals[i]);
    }
    putchar('\n');
}

void dumpDouble(ifd_entry *entry, header &tif)
{
    double * vals = (double*)(tif[entry->offset]);
    for (uint16_t i = 0; i>>1 < entry->count; i+=2) {
        printf("%f ", vals[i]);
    }
    putchar('\n');
}

void dumpASCII(ifd_entry *entry, header &tif)
{
    char * str;
    if (entry->count > 4) {
        str = (char *)(tif[entry->offset]);
    }
    else {
        str = (char*)&entry->offset;
    }

    iprintf("%*.*s\n", entry->count, entry->count, str);
}


void dumpIFD(ifd_entry *entry, header *tif)
{
    iprintf("    Addr: %p\n", entry);
    iprintf("    Tag: %s - 0x%04X\n", GetTag((ifd_tag)entry->tag), entry->tag);
    iprintf("    Type: %s - 0x%04X\n", GetType((ifd_type)entry->type), entry->type);
    iprintf("    Count: %u\n", entry->count);
    iprintf("    Value: ");
    switch(entry->type) {
    case BYTE:      { dumpByte(entry, *tif); break; }
    case ASCII:     { dumpASCII(entry, *tif); break; }
    case SHORT:     { dumpShort(entry, *tif); break; }
    case LONG:      { dumpLong(entry, *tif); break; }
    case RATIONAL:  { dumpRational(entry, *tif); break; }
    case SBYTE:     { dumpSShort(entry, *tif); break; }
    case UNDEFINED: { dumpUndef(entry, *tif); break; }
    case SSHORT:    { dumpSShort(entry, *tif); break; }
    case SLONG:     { dumpSLong(entry, *tif); break; }
    case SRATIONAL: { dumpSRational(entry, *tif); break; }
    case FLOAT:     { dumpFloat(entry, *tif); break; }
    case DOUBLE:    { dumpDouble(entry, *tif); break; }
    case TYPE_16BIT: { break; };
    }
}

void dumpTiff(tiff::header *tif)
{
    iprintf("TIFF: %p\n", tif);
    iprintf("order:  0x%04X\n", tif->order);
    iprintf("42:     0x%04X\n", tif->fortytwo);
    iprintf("offset: 0x%08X\n", tif->offset);

    ifd_struct *ifd = tif->GetIFD();
    iprintf("  IFD: %p - Count: %u\n", ifd, ifd->count);
    uint16_t count = ifd->count;
    ifd_entry *entry = ifd->entry;
    for (uint16_t i = 0; i < count; i++) {
        iprintf("  IFD[%d]:\n", i);
        dumpIFD(entry, tif);
        entry = entry->GetNext();
    }
}

}

extern const unsigned long lena_buf_len;
extern const unsigned char lena_buf[];


static http_gethandler *oldhand;
void ServeLena(int fd)
{
   writestring( fd,
         "HTTP/1.0 200 OK\r\nPragma: no-cache\r\nContent-Type: image/jpeg\r\n\r\n" );
    jpge::params comp_params;
    compress_image_to_fd(fd, 512, 512, 3, lena_dat, comp_params);
}

/*-------------------------------------------------------------------
 * By creating a HTTP GET handler, and registering it with
 * SetNewGetHandler( MyDoGet ), this function will be called in
 * response to a HTTP GET request instead of the system GET handler.
 * ----------------------------------------------------------------*/
int MyDoGet( int sock, PSTR url, PSTR rxBuffer )
{
    if ( ( url[0] != 0 ) && ( httpstricmp( url, "LENA.JPG" ) == 1 ) ) {
        ServeLena(sock);
        return 1;
    }

    return ( *oldhand ) ( sock, url, rxBuffer ); // Password was acceptable.
}

void UserMain( void * pd )
{
    initWithWeb();
    EnableSmartTraps();


    iprintf("Application: %s\r\nNNDK Revision: %s\r\n",AppName,GetReleaseTag());

    getchar();

    iprintf("\n\n  -----  DUMPING LENA  -----  \n\n");
    tiff::Tiff tif;
    tif.head = (tiff::header *)lena_buf;
    dumpTiff( tif.head );

    uint32_t offset = tif.GetEntry(tiff::STRIP_OFFSETS)->offset;
    lena_dat = (*tif.head)[offset];

    iprintf("lena_dat: %p\n", lena_dat);
    oldhand = SetNewGetHandler( MyDoGet );

    while(1) {
        OSTimeDly(TICKS_PER_SECOND);
    }
}
